/*
 * This file implements the Benchy firmware application.
 */

#include <cyu3system.h>
#include <cyu3os.h>
#include <cyu3error.h>
#include <cyu3uart.h>
#include <cyu3dma.h>
#include <cyu3usb.h>
#include <uart_regs.h>
#include <cyu3utils.h>
#include "BenchyFx3.h"
#include "DebugConsole.h"
#include "cyu3gpif.h"
#include "cyu3pib.h"
#include "pib_regs.h"
#include <cyu3gpio.h>

/* This file should be included only once as it contains
 * structure definitions. Including it in multiple places
 * can result in linker error. */

#include "cyfxgpif2config.h"

static CyU3PThread       AppThread;

CyU3PDmaChannel glChHandleSlFifoUtoP;   /* DMA Channel handle for U2P transfer. */
CyU3PDmaChannel glChHandleSlFifoPtoU;   /* DMA Channel handle for P2U transfer. */

CyU3PDmaChannel   glChHandleUsbtoUart;          /* DMA AUTO (USB TO UART) channel handle.*/
CyU3PDmaChannel   glChHandleUarttoUsb;          /* DMA AUTO_SIG(UART TO USB) channel handle.*/
CyBool_t          glIsApplnActive = CyFalse;    /* Whether the application is active or not. */
CyU3PUartConfig_t glUartConfig = {0};           /* Current UART configuration. */
volatile uint16_t glPktsPending = 0;            /* Number of packets that have been committed since last check. */

/* CDC Class specific requests to be handled by this application. */
#define SET_LINE_CODING        0x20
#define GET_LINE_CODING        0x21
#define SET_CONTROL_LINE_STATE 0x22

void
CyFxAppErrorHandler ( CyU3PReturnStatus_t apiRetStatus    /* API return status */ )
{
	/* Application failed with the error code apiRetStatus */

	/* Add custom debug or recovery actions here */

	/* Loop Indefinitely */
	for (;;)
	{
		/* Thread sleep : 100 ms */
		CyU3PThreadSleep (100);
	}
}

void DmaCallback(
		CyU3PDmaChannel   *chHandle, /* Handle to the DMA channel. */
		CyU3PDmaCbType_t   type,     /* Callback type.             */
		CyU3PDmaCBInput_t *input)    /* Callback status.           */
{
	if (type == CY_U3P_DMA_CB_PROD_EVENT)
	{
		CyU3PDmaChannelCommitBuffer (&glChHandleUarttoUsb, input->buffer_p.count, 0);
		glPktsPending++;
	}
}

/* This function starts the USBUART application */
void AppStart( void )
{ 
	uint16_t size = 0;
	CyU3PEpConfig_t epCfg;
	CyU3PDmaChannelConfig_t dmaCfg;
	CyU3PReturnStatus_t apiRetStatus = CY_U3P_SUCCESS;
	CyU3PUSBSpeed_t usbSpeed = CyU3PUsbGetSpeed();
	uint8_t burstLength = 0;

	/* Based on the Bus speed configure the endpoint packet size */
	switch (usbSpeed)
	{
	case CY_U3P_FULL_SPEED:
		size = 64;
		break;

	case CY_U3P_HIGH_SPEED:
		size = 512;
		break;

	case  CY_U3P_SUPER_SPEED:
		/* Turning low power mode off to avoid USB transfer delays. */
		CyU3PUsbLPMDisable ();
		size = 1024;
		break;

	default:
		CyFxAppErrorHandler (CY_U3P_ERROR_FAILURE);
		break;
	}

	/////////////////// Virtual COM port to FPGA register UART ///////////////////////

	CyU3PMemSet ((uint8_t *)&epCfg, 0, sizeof (epCfg));
	epCfg.enable = CyTrue;
	epCfg.epType = CY_U3P_USB_EP_BULK;
	epCfg.burstLen = 16;
	epCfg.streams = 0;
	epCfg.pcktSize = size;

	/* Producer endpoint configuration */
	apiRetStatus = CyU3PSetEpConfig(CY_FX_EP_PRODUCER_FPGA, &epCfg);
	CheckStatus("CyU3PSetEpConfig(CY_FX_EP_PRODUCER_FPGA)", apiRetStatus);

	/* Consumer endpoint configuration */
	apiRetStatus = CyU3PSetEpConfig(CY_FX_EP_CONSUMER_FPGA, &epCfg);
	CheckStatus("CyU3PSetEpConfig(CY_FX_EP_CONSUMER_FPGA)", apiRetStatus);

	/* Interrupt endpoint configuration */
	epCfg.epType = CY_U3P_USB_EP_INTR;
	epCfg.pcktSize = 64;
	epCfg.isoPkts = 1;

	apiRetStatus = CyU3PSetEpConfig(CY_FX_EP_INTERRUPT_FPGA, &epCfg);
	CheckStatus("CyU3PSetEpConfig(CY_FX_EP_INTERRUPT_FPGA)", apiRetStatus);

	/* Create a DMA_AUTO channel between usb producer socket and uart consumer socket */
	dmaCfg.size = size;
	dmaCfg.count = CY_FX_USBUART_DMA_BUF_COUNT;
	dmaCfg.prodSckId = CY_FX_EP_PRODUCER_FPGA_SOCKET;
	dmaCfg.consSckId = CY_U3P_LPP_SOCKET_UART_CONS ;
	dmaCfg.dmaMode = CY_U3P_DMA_MODE_BYTE;
	dmaCfg.notification = 0;
	dmaCfg.cb = NULL;
	dmaCfg.prodHeader = 0;
	dmaCfg.prodFooter = 0;
	dmaCfg.consHeader = 0;
	dmaCfg.prodAvailCount = 0;

	apiRetStatus = CyU3PDmaChannelCreate(&glChHandleUsbtoUart, CY_U3P_DMA_TYPE_AUTO, &dmaCfg);
	CheckStatus("CyU3PDmaChannelCreate(&glChHandleUsbtoUart)", apiRetStatus);

	/* Create a DMA_MANUAL channel between uart producer socket and usb consumer socket */
	/* Use a smaller buffer size (32 bytes) to ensure that packets get filled in a short time. */
	dmaCfg.size         = 32;
	dmaCfg.prodSckId    = CY_U3P_LPP_SOCKET_UART_PROD;
	dmaCfg.consSckId    = CY_FX_EP_CONSUMER_FPGA_SOCKET;
	dmaCfg.notification = CY_U3P_DMA_CB_PROD_EVENT;
	dmaCfg.cb           = DmaCallback;

	apiRetStatus = CyU3PDmaChannelCreate(&glChHandleUarttoUsb, CY_U3P_DMA_TYPE_MANUAL, &dmaCfg);
	CheckStatus("CyU3PDmaChannelCreate(&glChHandleUarttoUsb)", apiRetStatus);

	/* Set DMA Channel transfer size */
	apiRetStatus = CyU3PDmaChannelSetXfer(&glChHandleUsbtoUart,0);
	CheckStatus("CyU3PDmaChannelSetXfer(&glChHandleUsbtoUart)", apiRetStatus);

	apiRetStatus = CyU3PDmaChannelSetXfer(&glChHandleUarttoUsb, 0);
	CheckStatus("CyU3PDmaChannelSetXfer(&glChHandleUarttoUsb)", apiRetStatus);

	/////////////////// Slave FIFO data ///////////////////////

	CyU3PMemSet ((uint8_t *)&epCfg, 0, sizeof (epCfg));
    epCfg.enable = CyTrue;
    epCfg.epType = CY_U3P_USB_EP_BULK;
#ifdef STREAM_IN_OUT
    epCfg.burstLen = burstLength;
#else
    epCfg.burstLen = 1;
#endif
    epCfg.streams = 0;
    epCfg.pcktSize = size;

    /* Producer endpoint configuration */
    apiRetStatus = CyU3PSetEpConfig(CY_FX_EP_PRODUCER_DATA, &epCfg);
    CheckStatus("CyU3PSetEpConfig(CY_FX_EP_PRODUCER_DATA)", apiRetStatus);

    /* Consumer endpoint configuration */
    apiRetStatus = CyU3PSetEpConfig(CY_FX_EP_CONSUMER_DATA, &epCfg);
    CheckStatus("CyU3PSetEpConfig(CY_FX_EP_CONSUMER_DATA)", apiRetStatus);

    /* Create a DMA AUTO channel for U2P transfer.
       DMA size is set based on the USB speed. */
    dmaCfg.size  = DMA_BUF_SIZE* size;
    dmaCfg.count = CY_FX_SLFIFO_DMA_BUF_COUNT_U_2_P;
	dmaCfg.prodSckId = CY_FX_PRODUCER_DATA_SOCKET;
    dmaCfg.consSckId = CY_FX_CONSUMER_PPORT_SOCKET;
    dmaCfg.dmaMode = CY_U3P_DMA_MODE_BYTE;
    /* Enabling the callback for produce event. */
    dmaCfg.notification = 0;
    dmaCfg.cb = NULL;
    dmaCfg.prodHeader = 0;
    dmaCfg.prodFooter = 0;
    dmaCfg.consHeader = 0;
    dmaCfg.prodAvailCount = 0;

    apiRetStatus = CyU3PDmaChannelCreate (&glChHandleSlFifoUtoP, CY_U3P_DMA_TYPE_AUTO, &dmaCfg);
    CheckStatus("CyU3PDmaChannelCreate (&glChHandleSlFifoUtoP)", apiRetStatus);

    /* Create a DMA AUTO channel for P2U transfer. */
    dmaCfg.size  = DMA_BUF_SIZE*size; //increase buffer size for higher performance
    dmaCfg.count = CY_FX_SLFIFO_DMA_BUF_COUNT_P_2_U; // increase buffer count for higher performance
    dmaCfg.prodSckId = CY_FX_PRODUCER_PPORT_SOCKET;
    dmaCfg.consSckId = CY_FX_CONSUMER_DATA_SOCKET;
    dmaCfg.cb = NULL;
    apiRetStatus = CyU3PDmaChannelCreate (&glChHandleSlFifoPtoU, CY_U3P_DMA_TYPE_AUTO, &dmaCfg);
    CheckStatus("CyU3PDmaChannelCreate (&glChHandleSlFifoPtoU)", apiRetStatus);

    /* Flush the Endpoint memory */
    CyU3PUsbFlushEp(CY_FX_EP_PRODUCER_DATA);
    CyU3PUsbFlushEp(CY_FX_EP_CONSUMER_DATA);

    /* Set DMA channel transfer size. */
    apiRetStatus = CyU3PDmaChannelSetXfer (&glChHandleSlFifoUtoP, CY_FX_SLFIFO_DMA_TX_SIZE);
    CheckStatus("CyU3PDmaChannelSetXfer (&glChHandleSlFifoUtoP)", apiRetStatus);

    apiRetStatus = CyU3PDmaChannelSetXfer (&glChHandleSlFifoPtoU, CY_FX_SLFIFO_DMA_RX_SIZE);
    CheckStatus("CyU3PDmaChannelSetXfer (&glChHandleSlFifoPtoU)", apiRetStatus);

    /* Update the status flag. */
    CyU3PDebugPrint (4, "Active\r\n", apiRetStatus);
    glIsApplnActive = CyTrue;
    CyU3PGpioSetValue (59, CyFalse);

	/* Update the status flag. */
	glIsApplnActive = CyTrue;
} 

void AppStop ( void)
{
	CyU3PEpConfig_t epCfg;
	CyU3PReturnStatus_t apiRetStatus = CY_U3P_SUCCESS;

	CyU3PDebugPrint(4, "AppStop");

	/* Update the flag. */
	glIsApplnActive = CyFalse;

	/////////////////// Virtual COM port to FPGA register UART ///////////////////////

	/* Flush the endpoint memory */
	CyU3PUsbFlushEp(CY_FX_EP_PRODUCER_FPGA);
	CyU3PUsbFlushEp(CY_FX_EP_CONSUMER_FPGA);
	CyU3PUsbFlushEp(CY_FX_EP_INTERRUPT_FPGA);

	/* Destroy the channel */
	CyU3PDmaChannelDestroy (&glChHandleUsbtoUart);
	CyU3PDmaChannelDestroy (&glChHandleUarttoUsb);

	/* Disable endpoints. */
	CyU3PMemSet ((uint8_t *)&epCfg, 0, sizeof (epCfg));
	epCfg.enable = CyFalse;

	/* Producer endpoint configuration. */
	apiRetStatus = CyU3PSetEpConfig(CY_FX_EP_PRODUCER_FPGA, &epCfg);
	CheckStatus("CyU3PSetEpConfig(CY_FX_EP_PRODUCER_FPGA)", apiRetStatus);

	/* Consumer endpoint configuration. */
	apiRetStatus = CyU3PSetEpConfig(CY_FX_EP_CONSUMER_FPGA, &epCfg);
	CheckStatus("CyU3PSetEpConfig(CY_FX_EP_CONSUMER_FPGA)", apiRetStatus);

	/* Interrupt endpoint configuration. */
	apiRetStatus = CyU3PSetEpConfig(CY_FX_EP_INTERRUPT_FPGA, &epCfg);
	CheckStatus("CyU3PSetEpConfig(CY_FX_EP_INTERRUPT_FPGA)", apiRetStatus);

	/////////////////// Slave FIFO data ///////////////////////

    /* Flush the endpoint memory */
    CyU3PUsbFlushEp(CY_FX_EP_PRODUCER_DATA);
    CyU3PUsbFlushEp(CY_FX_EP_CONSUMER_DATA);

    /* Destroy the channel */
    CyU3PDmaChannelDestroy (&glChHandleSlFifoUtoP);
    CyU3PDmaChannelDestroy (&glChHandleSlFifoPtoU);

    /* Disable endpoints. */
    CyU3PMemSet ((uint8_t *)&epCfg, 0, sizeof (epCfg));
    epCfg.enable = CyFalse;

    /* Producer endpoint configuration. */
    apiRetStatus = CyU3PSetEpConfig(CY_FX_EP_PRODUCER_DATA, &epCfg);
    CheckStatus("CyU3PSetEpConfig(CY_FX_EP_PRODUCER_DATA)", apiRetStatus);

    /* Consumer endpoint configuration. */
    apiRetStatus = CyU3PSetEpConfig(CY_FX_EP_CONSUMER_DATA, &epCfg);
    CheckStatus("CyU3PSetEpConfig(CY_FX_EP_CONSUMER_DATA)", apiRetStatus);

}

/* This is the callback function to handle the USB events. */
void AppUSBEventCB ( CyU3PUsbEventType_t evtype, /* Event type */ uint16_t evdata  /* Event data */ )
{
	switch (evtype)
	{
	case CY_U3P_USB_EVENT_SETCONF:
		/* Stop the application before re-starting. */
		if (glIsApplnActive)
		{
			AppStop ();
		}
		/* Start the loop back function. */
		AppStart ();
		break;

	case CY_U3P_USB_EVENT_RESET:
	case CY_U3P_USB_EVENT_CONNECT:
	case CY_U3P_USB_EVENT_DISCONNECT:
		/* Stop the loop back function. */
		if (glIsApplnActive)
		{
			CyU3PUsbLPMEnable ();
			AppStop ();
		}
		break;

	default:
		break;
	}
}

/* Callback to handle the USB Setup Requests and CDC Class events */
static CyBool_t AppUSBSetupCB ( uint32_t setupdat0, /* SETUP Data 0 */ uint32_t setupdat1  /* SETUP Data 1 */ )
{
	uint16_t readCount = 0;
	uint8_t  bRequest, bReqType;
	uint8_t  bType, bTarget;
	uint16_t wValue, wIndex;
	uint8_t config_data[7];
	CyBool_t isHandled = CyFalse;
	CyU3PReturnStatus_t status = CY_U3P_SUCCESS;
	CyU3PReturnStatus_t apiRetStatus = CY_U3P_SUCCESS;
	CyU3PUartConfig_t uartConfig;

	/* Fast enumeration is used. Only requests addressed to the interface, class,
	 * vendor and unknown control requests are received by this function. */

	/* Decode the fields from the setup request. */
	bReqType = (setupdat0 & CY_U3P_USB_REQUEST_TYPE_MASK);
	bType    = (bReqType & CY_U3P_USB_TYPE_MASK);
	bTarget  = (bReqType & CY_U3P_USB_TARGET_MASK);
	bRequest = ((setupdat0 & CY_U3P_USB_REQUEST_MASK) >> CY_U3P_USB_REQUEST_POS);
	wValue   = ((setupdat0 & CY_U3P_USB_VALUE_MASK)   >> CY_U3P_USB_VALUE_POS);
    wIndex   = ((setupdat1 & CY_U3P_USB_INDEX_MASK)   >> CY_U3P_USB_INDEX_POS);

	if (bType == CY_U3P_USB_STANDARD_RQT)
	{
		/* Handle SET_FEATURE(FUNCTION_SUSPEND) and CLEAR_FEATURE(FUNCTION_SUSPEND)
		 * requests here. It should be allowed to pass if the device is in configured
		 * state and failed otherwise. */
		if ((bTarget == CY_U3P_USB_TARGET_INTF) && ((bRequest == CY_U3P_USB_SC_SET_FEATURE)
				|| (bRequest == CY_U3P_USB_SC_CLEAR_FEATURE)) && (wValue == 0))
		{
			if (glIsApplnActive)
				CyU3PUsbAckSetup ();
			else
				CyU3PUsbStall (0, CyTrue, CyFalse);

			isHandled = CyTrue;
		}
	}

    /* CLEAR_FEATURE request for endpoint is always passed to the setup callback
     * regardless of the enumeration model used. When a clear feature is received,
     * the previous transfer has to be flushed and cleaned up. This is done at the
     * protocol level. Since this is just a loopback operation, there is no higher
     * level protocol. So flush the EP memory and reset the DMA channel associated
     * with it. If there are more than one EP associated with the channel reset both
     * the EPs. The endpoint stall and toggle / sequence number is also expected to be
     * reset. Return CyFalse to make the library clear the stall and reset the endpoint
     * toggle. Or invoke the CyU3PUsbStall (ep, CyFalse, CyTrue) and return CyTrue.
     * Here we are clearing the stall. */
    if ((bTarget == CY_U3P_USB_TARGET_ENDPT) && (bRequest == CY_U3P_USB_SC_CLEAR_FEATURE)
            && (wValue == CY_U3P_USBX_FS_EP_HALT))
    {
        if (glIsApplnActive)
        {
            if (wIndex == CY_FX_EP_PRODUCER_DATA)
            {
                CyU3PDmaChannelReset (&glChHandleSlFifoUtoP);
                CyU3PUsbFlushEp(CY_FX_EP_PRODUCER_DATA);
                CyU3PUsbResetEp (CY_FX_EP_PRODUCER_DATA);
                CyU3PDmaChannelSetXfer (&glChHandleSlFifoUtoP, CY_FX_SLFIFO_DMA_TX_SIZE);
            }

            if (wIndex == CY_FX_EP_CONSUMER_DATA)
            {
                CyU3PDmaChannelReset (&glChHandleSlFifoPtoU);
                CyU3PUsbFlushEp(CY_FX_EP_CONSUMER_DATA);
                CyU3PUsbResetEp (CY_FX_EP_CONSUMER_DATA);
                CyU3PDmaChannelSetXfer (&glChHandleSlFifoPtoU, CY_FX_SLFIFO_DMA_RX_SIZE);
            }

            CyU3PUsbStall (wIndex, CyFalse, CyTrue);
            isHandled = CyTrue;
        }
    }

    /* Check for CDC Class Requests */
	if (bType == CY_U3P_USB_CLASS_RQT)
	{
		isHandled = CyTrue;

		/* CDC Specific Requests */
		/* set_line_coding */
		if (bRequest == SET_LINE_CODING)
		{
			status = CyU3PUsbGetEP0Data(0x07, config_data, &readCount);
			if (status != CY_U3P_SUCCESS)
			{
				CyFxAppErrorHandler(status);
			}
			if (readCount != 0x07)
			{
				CyFxAppErrorHandler(CY_U3P_ERROR_BAD_SIZE);
			}
			else
			{
				CyU3PMemSet ((uint8_t *)&uartConfig, 0, sizeof (uartConfig));
				uartConfig.baudRate = (CyU3PUartBaudrate_t)(config_data[0] | (config_data[1]<<8)|
						(config_data[2]<<16)|(config_data[3]<<24));
				if (config_data[4] == 0)
				{
					uartConfig.stopBit = CY_U3P_UART_ONE_STOP_BIT;
				}
				else if (config_data[4] == 2)
				{
					uartConfig.stopBit = CY_U3P_UART_TWO_STOP_BIT;
				}
				else
				{
					/* Give invalid value. */
					uartConfig.stopBit = (CyU3PUartStopBit_t)0;
				}
				if (config_data[5] == 1)
				{
					uartConfig.parity = CY_U3P_UART_ODD_PARITY;
				}
				else if (config_data[5] == 2)
				{
					uartConfig.parity = CY_U3P_UART_EVEN_PARITY;
				}
				else
				{
					/* 0 = no parity; any other value - invalid parity. */
					uartConfig.parity = CY_U3P_UART_NO_PARITY;
				}

				uartConfig.txEnable = CyTrue;
				uartConfig.rxEnable = CyTrue;
				uartConfig.flowCtrl = CyFalse;
				uartConfig.isDma = CyTrue;

				/* Set the uart configuration */
				apiRetStatus = CyU3PUartSetConfig (&uartConfig, NULL);
				if (apiRetStatus == CY_U3P_SUCCESS)
				{
					CyU3PMemCopy ((uint8_t *)&glUartConfig, (uint8_t *)&uartConfig,
							sizeof (CyU3PUartConfig_t));
				}
			}
		}
		/* get_line_coding */
		else if (bRequest == GET_LINE_CODING )
		{
			/* get current uart config */
			config_data[0] = glUartConfig.baudRate&(0x000000FF);
			config_data[1] = ((glUartConfig.baudRate&(0x0000FF00))>> 8);
			config_data[2] = ((glUartConfig.baudRate&(0x00FF0000))>>16);
			config_data[3] = ((glUartConfig.baudRate&(0xFF000000))>>24);
			if (glUartConfig.stopBit == CY_U3P_UART_ONE_STOP_BIT)
			{
				config_data[4] = 0;
			}
			else /* CY_U3P_UART_TWO_STOP_BIT */
			{
				config_data[4] = 2;
			}

			if (glUartConfig.parity == CY_U3P_UART_EVEN_PARITY)
			{
				config_data[5] = 2;
			}
			else if (glUartConfig.parity == CY_U3P_UART_ODD_PARITY)
			{
				config_data[5] = 1;
			}
			else
			{
				config_data[5] = 0;
			}
			config_data[6] =  0x08;
			status = CyU3PUsbSendEP0Data( 0x07, config_data);
			if (status != CY_U3P_SUCCESS)
			{
				CyFxAppErrorHandler(status);
			}
		}
		/* SET_CONTROL_LINE_STATE */
		else if (bRequest == SET_CONTROL_LINE_STATE)
		{
			if (glIsApplnActive)
			{
				CyU3PUsbAckSetup ();
			}
			else
				CyU3PUsbStall (0, CyTrue, CyFalse);
		}
		else
		{
			status = CY_U3P_ERROR_FAILURE;
		}

		if (status != CY_U3P_SUCCESS)
		{
			isHandled = CyFalse;
		}
	}

	return isHandled;
}

CyBool_t AppLPMRqtCB ( CyU3PUsbLinkPowerMode link_mode)
{
	return CyTrue;
}

void gpif_error_cb(CyU3PPibIntrType cbType, uint16_t cbArg)
{

	if(cbType==CYU3P_PIB_INTR_ERROR)
	{
		switch (CYU3P_GET_PIB_ERROR_TYPE(cbArg))
		{
		case CYU3P_PIB_ERR_THR0_WR_OVERRUN:
			CyU3PDebugPrint (4, "CYU3P_PIB_ERR_THR0_WR_OVERRUN");
			break;
		case CYU3P_PIB_ERR_THR1_WR_OVERRUN:
			CyU3PDebugPrint (4, "CYU3P_PIB_ERR_THR1_WR_OVERRUN");
			break;
		case CYU3P_PIB_ERR_THR2_WR_OVERRUN:
			CyU3PDebugPrint (4, "CYU3P_PIB_ERR_THR2_WR_OVERRUN");
			break;
		case CYU3P_PIB_ERR_THR3_WR_OVERRUN:
			CyU3PDebugPrint (4, "CYU3P_PIB_ERR_THR3_WR_OVERRUN");
			break;

		case CYU3P_PIB_ERR_THR0_RD_UNDERRUN:
			CyU3PDebugPrint (4, "CYU3P_PIB_ERR_THR0_RD_UNDERRUN");
			break;
		case CYU3P_PIB_ERR_THR1_RD_UNDERRUN:
			CyU3PDebugPrint (4, "CYU3P_PIB_ERR_THR1_RD_UNDERRUN");
			break;
		case CYU3P_PIB_ERR_THR2_RD_UNDERRUN:
			CyU3PDebugPrint (4, "CYU3P_PIB_ERR_THR2_RD_UNDERRUN");
			break;
		case CYU3P_PIB_ERR_THR3_RD_UNDERRUN:
			CyU3PDebugPrint (4, "CYU3P_PIB_ERR_THR3_RD_UNDERRUN");
			break;

		default:
			CyU3PDebugPrint (4, "No Error :%d\n ",CYU3P_GET_PIB_ERROR_TYPE(cbArg));
			break;
		}
	}
}

/* This function initializes the USB module, UART module and sets the enumeration descriptors */
void AppInit(void)

{
    CyU3PPibClock_t pibClock;
    CyU3PGpioClock_t gpioClock;
    CyU3PGpioSimpleConfig_t gpioConfig;
	CyU3PReturnStatus_t apiRetStatus = CY_U3P_SUCCESS;

	/* Start the USB functionality. */
	apiRetStatus = CyU3PUsbStart();
	CheckStatus("CyU3PUsbStart", apiRetStatus);

	/////////////////// Virtual COM port to FPGA register UART ///////////////////////

	/* Initialize the UART module */
	apiRetStatus = CyU3PUartInit ();
	CheckStatus("CyU3PUartInit", apiRetStatus);

	/* Configure the UART */
	CyU3PMemSet ((uint8_t *)&glUartConfig, 0, sizeof (glUartConfig));
	glUartConfig.baudRate = CY_U3P_UART_BAUDRATE_115200;
	glUartConfig.stopBit = CY_U3P_UART_ONE_STOP_BIT;
	glUartConfig.parity = CY_U3P_UART_NO_PARITY;
	glUartConfig.flowCtrl = CyFalse;
	glUartConfig.txEnable = CyTrue;
	glUartConfig.rxEnable = CyTrue;
	glUartConfig.isDma = CyTrue;

	/* Set the UART configuration */
	apiRetStatus = CyU3PUartSetConfig (&glUartConfig, NULL);
	CheckStatus("CyU3PUartSetConfig(&glUartConfig)", apiRetStatus);

	/////////////////// Slave FIFO data ///////////////////////

    /* Initialize the p-port block. */
    pibClock.clkDiv = 2;
    pibClock.clkSrc = CY_U3P_SYS_CLK;
    pibClock.isHalfDiv = CyFalse;
    /* Disable DLL for sync GPIF */
    pibClock.isDllEnable = CyFalse;
    apiRetStatus = CyU3PPibInit(CyTrue, &pibClock);
    CheckStatus("CyU3PPibInit(CyTrue, &pibClock)", apiRetStatus);


    /* Load the GPIF configuration for Slave FIFO sync mode. */
    apiRetStatus = CyU3PGpifLoad (&CyFxGpifConfig);
    CheckStatus("CyU3PGpifLoad (&CyFxGpifConfig)", apiRetStatus);

    CyU3PGpifSocketConfigure (0,CY_FX_PRODUCER_PPORT_SOCKET,6,CyFalse,1);

    CyU3PGpifSocketConfigure (3,CY_FX_CONSUMER_PPORT_SOCKET,1,CyFalse,1);

    /* Start the state machine. */
    apiRetStatus = CyU3PGpifSMStart (RESET,ALPHA_RESET);
    CheckStatus("CyU3PGpifSMStart (RESET,ALPHA_RESET)", apiRetStatus);

    /* Init the GPIO module */
    CyU3PMemSet ((uint8_t *)&gpioClock, 0, sizeof (gpioClock));
	gpioClock.fastClkDiv = 2;
	gpioClock.slowClkDiv = 0;
	gpioClock.simpleDiv = CY_U3P_GPIO_SIMPLE_DIV_BY_2;
	gpioClock.clkSrc = CY_U3P_SYS_CLK;
	gpioClock.halfDiv = 0;
	apiRetStatus = CyU3PGpioInit(&gpioClock, NULL);
	CheckStatus("CyU3PGpioInit(&gpioClock)", apiRetStatus);

    /* Configure GPIO 59 as output */
	CyU3PMemSet ((uint8_t *)&gpioConfig, 0, sizeof (gpioConfig));
    gpioConfig.outValue = CyTrue;
    gpioConfig.driveLowEn = CyTrue;
    gpioConfig.driveHighEn = CyTrue;
    gpioConfig.inputEn = CyFalse;
    gpioConfig.intrMode = CY_U3P_GPIO_NO_INTR;
    apiRetStatus = CyU3PGpioSetSimpleConfig(59, &gpioConfig);
	CheckStatus("CyU3PGpioSetSimpleConfig(59, &gpioConfig)", apiRetStatus);
//    CyU3PDeviceGpioOverride(54, CyTrue);
//    apiRetStatus = CyU3PGpioSetSimpleConfig(54, &gpioConfig);

    /* callback to see if there is any overflow of data on the GPIF II side*/
    CyU3PPibRegisterCallback(gpif_error_cb,0xffff);

    ////////////////// register callbacks ////////////////////////////////

	/* Setup the callback to handle the setup requests */
	CyU3PUsbRegisterSetupCallback(AppUSBSetupCB, CyTrue);

	/* Setup the callback to handle the USB events. */
	CyU3PUsbRegisterEventCallback(AppUSBEventCB);

	/* Register a callback to handle LPM requests from the USB 3.0 host. */
	CyU3PUsbRegisterLPMRequestCallback(AppLPMRqtCB);

	/* Set the USB enumeration descriptors */

	/* Super speed device descriptor. */
	apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_SS_DEVICE_DESCR, 0, (uint8_t *)CyFxUSB30DeviceDscr);
	CheckStatus("CyU3PUsbSetDesc(CY_U3P_USB_SET_SS_DEVICE_DESCR)", apiRetStatus);

	/* High speed device descriptor. */
	apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_HS_DEVICE_DESCR, 0, (uint8_t *)CyFxUSB20DeviceDscr);
	CheckStatus("CyU3PUsbSetDesc(CY_U3P_USB_SET_HS_DEVICE_DESCR)", apiRetStatus);

	/* BOS descriptor */
	apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_SS_BOS_DESCR, 0, (uint8_t *)CyFxUSBBOSDscr);
	CheckStatus("CyU3PUsbSetDesc(CY_U3P_USB_SET_SS_BOS_DESCR)", apiRetStatus);

	/* Device qualifier descriptor */
	apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_DEVQUAL_DESCR, 0, (uint8_t *)CyFxUSBDeviceQualDscr);
	CheckStatus("CyU3PUsbSetDesc(CY_U3P_USB_SET_DEVQUAL_DESCR)", apiRetStatus);

	/* Super speed configuration descriptor */
	apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_SS_CONFIG_DESCR, 0, (uint8_t *)CyFxUSBSSConfigDscr);
	CheckStatus(")", apiRetStatus);

	/* High speed configuration descriptor*/
	apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_HS_CONFIG_DESCR, 0, (uint8_t *)CyFxUSBHSConfigDscr);
	CheckStatus("CyU3PUsbSetDesc(CY_U3P_USB_SET_HS_CONFIG_DESCR)", apiRetStatus);

	/* Full speed configuration descriptor */
	apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_FS_CONFIG_DESCR, 0, (uint8_t *)CyFxUSBHSConfigDscr);
	CheckStatus("CyU3PUsbSetDesc(CY_U3P_USB_SET_FS_CONFIG_DESCR)", apiRetStatus);

	/* String descriptor 0 */
	apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 0, (uint8_t *)CyFxUSBStringLangIDDscr);
	CheckStatus("CyU3PUsbSetDesc(CY_U3P_USB_SET_FS_CONFIG_DESCR)", apiRetStatus);

	/* String descriptor 1 */
	apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 1, (uint8_t *)CyFxUSBStringManufactureDscr);
	CheckStatus("CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 1)", apiRetStatus);

	/* String descriptor 2 */
	apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 2, (uint8_t *)CyFxUSBStringProductDscr);
	CheckStatus("CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 2)", apiRetStatus);

	/* String descriptor 3 */
	apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 3, (uint8_t *)CyFxUSBStringFpgaDscr);
	CheckStatus("CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 3)", apiRetStatus);

	/* String descriptor 4 */
	apiRetStatus = CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 4, (uint8_t *)CyFxUSBStringDataDscr);
	CheckStatus("CyU3PUsbSetDesc(CY_U3P_USB_SET_STRING_DESCR, 4)", apiRetStatus);

	/* Connect the USB Pins with super speed operation enabled. */
	apiRetStatus = CyU3PConnectState(CyTrue, CyTrue);
	CheckStatus("CyU3PConnectState", apiRetStatus);
}

/* Entry function for the USBUARTAppThread */
void AppThread_Entry ( uint32_t input)
{   
	uint32_t regValueEn = 0, regValueDs = 0;

	/* Initialize the Benchy firmware application. */
	AppInit();

    while (!glIsApplnActive)
    {
    	CyU3PThreadSleep(10);
    }

	InitializeUsbDebugConsole();

	/* UART Config Value for Enabling Rx Block */
	regValueEn = UART->lpp_uart_config;

	/* UART Config Value for Disabling the Rx Block  */
	regValueDs = UART->lpp_uart_config & (~(CY_U3P_LPP_UART_RTS | CY_U3P_LPP_UART_RX_ENABLE));

	for (;;)
	{
		if (glIsApplnActive)
		{
			/* While the application is active, check for data sent during the last 50 ms. If no data
               has been sent to the host, use the channel wrap-up feature to send any partial buffer to
               the USB host.
			 */
			if (glPktsPending == 0)
			{
				/* Disable UART Receiver Block */
				UART->lpp_uart_config = regValueDs;

				CyU3PDmaChannelSetWrapUp (&glChHandleUarttoUsb);

				/* Enable UART Receiver Block */
				UART->lpp_uart_config = regValueEn;
			}

			glPktsPending = 0;
		}

		CyU3PThreadSleep (50);
	}
}

/* Application define function known to the RTOS by name which creates the threads - Don't change the name of this function */
void CyFxApplicationDefine ( void )
{
	void *ptr = NULL;
	uint32_t retThrdCreate = CY_U3P_SUCCESS;

	/* Allocate the memory for the thread*/
	ptr = CyU3PMemAlloc (CY_FX_USBUART_THREAD_STACK);

	/* Create the thread for the application */
	retThrdCreate = CyU3PThreadCreate (&AppThread,          /* USBUART Example App Thread structure */
			"21:USBUART_DMA_mode",                   /* Thread ID and Thread name */
			AppThread_Entry,                  /* USBUART Example App Thread Entry function */
			0,                                       /* No input parameter to thread */
			ptr,                                     /* Pointer to the allocated thread stack */
			CY_FX_USBUART_THREAD_STACK,              /* USBUART Example App Thread stack size */
			CY_FX_USBUART_THREAD_PRIORITY,            /* USBUART Example App Thread priority */
			CY_FX_USBUART_THREAD_PRIORITY,            /* USBUART Example App Thread priority */
			CYU3P_NO_TIME_SLICE,                     /* No time slice for the application thread */
			CYU3P_AUTO_START                         /* Start the Thread immediately */
	);
	/* Check the return code */
	if (retThrdCreate != 0)
	{

		/* Loop indefinitely */
		while(1);
	}
}

/* Main function */
int main ( void)
{
	CyU3PIoMatrixConfig_t io_cfg;
	CyU3PReturnStatus_t status = CY_U3P_SUCCESS;
    CyU3PSysClockConfig_t clkCfg;

        /* setSysClk400 clock configurations */
        clkCfg.setSysClk400 = CyTrue;   /* FX3 device's master clock is set to a frequency > 400 MHz */
        clkCfg.cpuClkDiv = 2;           /* CPU clock divider */
        clkCfg.dmaClkDiv = 2;           /* DMA clock divider */
        clkCfg.mmioClkDiv = 2;          /* MMIO clock divider */
        clkCfg.useStandbyClk = CyFalse; /* device has no 32KHz clock supplied */
        clkCfg.clkSrc = CY_U3P_SYS_CLK; /* Clock source for a peripheral block  */


	/* Initialize the device */
//	status = CyU3PDeviceInit (0);
    status = CyU3PDeviceInit (&clkCfg);
	if (status != CY_U3P_SUCCESS)
	{
		goto handle_fatal_error;
	}

	/* Initialize the caches. Enable the instruction cache only. */
	status = CyU3PDeviceCacheControl (CyTrue, CyFalse, CyFalse);
	if (status != CY_U3P_SUCCESS)
	{
		goto handle_fatal_error;
	}

	/* Configure the IO matrix */
	CyU3PMemSet((uint8_t *)&io_cfg,0,sizeof(io_cfg));
//	io_cfg.isDQ32Bit = CyFalse;
	io_cfg.useUart   = CyTrue;
	io_cfg.useI2C    = CyFalse;
	io_cfg.useI2S    = CyFalse;
	io_cfg.useSpi    = CyFalse;
    io_cfg.isDQ32Bit = CyTrue;
    io_cfg.lppMode   = CY_U3P_IO_MATRIX_LPP_DEFAULT;
    io_cfg.gpioSimpleEn[0]  = 0;
    io_cfg.gpioSimpleEn[1]  = 0x08000000; /* GPIO 59 */
//    io_cfg.gpioSimpleEn[1]  = 0x00400000; /* GPIO 54 LED */
    io_cfg.gpioComplexEn[0] = 0;
    io_cfg.gpioComplexEn[1] = 0;//	io_cfg.lppMode   = CY_U3P_IO_MATRIX_LPP_UART_ONLY;
//	io_cfg.gpioSimpleEn[0]  = 0;
//	io_cfg.gpioSimpleEn[1]  = 0;
//	io_cfg.gpioComplexEn[0] = 0;
//	io_cfg.gpioComplexEn[1] = 0;
	status = CyU3PDeviceConfigureIOMatrix (&io_cfg);
	if (status != CY_U3P_SUCCESS)
	{
		goto handle_fatal_error;
	}

	/* Initialize the RTOS kernel */
	CyU3PKernelEntry ();

	return 0;

	handle_fatal_error:
	while(1);

}

